import math
from enum import Enum, auto
from typing import List, Dict
from abc import ABC, abstractmethod
from gp_framework.genotype import Genotype
from gp_framework.phenotype.phenotype import PhenotypeConverter
from gp_framework.exception import InvalidParameterException, InvalidApplicationException
from gp_framework.phenotype.number_generator import NumberGenerator, NumberGeneratorPhenotypeConverter
from gp_framework.phenotype.string_search import StringPhenotypeConverter
from gp_framework.phenotype.blackjack import Player, BlackJackTable, PlayerParameters, PlayerConverter


class Application(Enum):
    STRING_MATCH = auto()
    PRIMES = auto()
    BLACK_JACK = auto()


class PopulationFitnessCalculator(ABC):
    def __init__(self, phenotype_converter: PhenotypeConverter, target_fitness: int):
        self._converter = phenotype_converter
        self._target_fitness = target_fitness

    @property
    def target_fitness(self):
        return self._target_fitness

    def calculate_normalized_fitness(self, population: List[Genotype]) -> Dict[Genotype, float]:
        return {elem[0]: elem[1]/self._target_fitness for elem in self.calculate_fitness(population)}

    @abstractmethod
    def calculate_fitness(self, population: List[Genotype]) -> Dict[Genotype, int]:
        """
        Calculate the fitness of every genotype in the population. They will be automatically converted to a phenotype.
        :return: A dictionary mapping each genotype to an integer representing its fitness
        """
        pass


class FitnessCalculator(ABC):
    def __init__(self, phenotype_converter: PhenotypeConverter, application_arguments: List[any]):
        """
        If there is a known target fitness, it should be specified in the child's constructor
        :param application_arguments: This varies depending on the needs of the concrete FitnessCalculator
        """
        self._application_arguments = application_arguments
        self._target_fitness = -1
        self._converter = phenotype_converter

    @property
    def target_fitness(self):
        return self._target_fitness

    def calculate_normalized_fitness(self, genotype: Genotype) -> float:
        return self.calculate_fitness(genotype) / self._target_fitness

    @abstractmethod
    def calculate_fitness(self, genotype: Genotype) -> int:
        """
        Calculate the fitness of the given phenotype
        :param genotype: the Genotype to calculate the fitness of. It will be
        automatically converted to the correct phenotype.
        :return: the fitness of phenotype
        """
        pass

    @property
    def converter(self) -> PhenotypeConverter:
        return self._converter


class FitnessCalculatorStringMatch(FitnessCalculator):
    def __init__(self, phenotype_converter: PhenotypeConverter, application_arguments: List[any]):
        super().__init__(phenotype_converter, application_arguments)
        self._target_string = application_arguments[0]
        self._target_fitness = self.calculate_fitness_of_string(self._target_string)

    def _to_string(self, genotype: Genotype) -> str:
        return self._converter.convert(genotype)

    def calculate_fitness_of_string(self, string: str):
        if len(self._target_string) != len(string):
            return 0

        fitness = 0
        for i in range(len(string)):
            distance = abs(ord(self._target_string[i]) - ord(string[i]))
            fitness += 127 - distance
        return fitness

    def calculate_fitness(self, genotype: Genotype) -> int:
        """
        This is similar to the other function, but returns an unnormalized value
        """

        phenotype = self._to_string(genotype)
        return self.calculate_fitness_of_string(phenotype)

    @staticmethod
    def _normalize_ascii_value(value: int) -> int:
        # This can always be tweaked later if the target string is to include
        # some symbol outside the range [32, 126].
        """if 32 <= value <= 126:
            return value
        value = (abs(value) % (126 - 32)) + 32"""
        return value


class FitnessCalculatorPrimes(FitnessCalculator):
    def __init__(self, target_num_primes: int, phenotype_converter: NumberGeneratorPhenotypeConverter,
                 application_arguments: List[any]):
        super().__init__(phenotype_converter, application_arguments)
        self._length_of_list_of_primes = target_num_primes
        self._target_fitness = self._length_of_list_of_primes

    @staticmethod
    def _is_prime(x: int) -> bool:
        for i in range(2, math.ceil(math.sqrt(x))):
            if x % i == 0:
                return False
        return True

    def calculate_fitness(self, genotype: Genotype) -> int:
        prime_number_generator: NumberGenerator = self._converter.convert(genotype)
        prime_number_generator.make_list(self._length_of_list_of_primes)
        set_of_primes = {x for x in prime_number_generator.list_of_numbers}
        fitness = 0
        # check in the set so as not to double count anything
        for num in set_of_primes:
            if self._is_prime(num):
                fitness += 1
        return fitness


class PopulationFitnessCalculatorBlackJack(PopulationFitnessCalculator):
    def __init__(self, target_fitness, number_of_rounds):
        super().__init__(PlayerConverter(), target_fitness)
        self._number_of_rounds = number_of_rounds  # how many rounds the players play

    def calculate_fitness(self, population: List[Genotype]) -> Dict[Genotype, int]:
        players = []
        for genotype in population:
            players.append(self._converter.convert(genotype))
        table = BlackJackTable(players)
        table.play_rounds(self._number_of_rounds)
        fitness_dict = {}
        for i in range(len(players)):
            fitness_dict[population[i]] = table.players[i].money
        return fitness_dict


class FitnessCalculatorBlackJack(FitnessCalculator):
    def __init__(self, phenotype_converter: PhenotypeConverter, application_arguments: List[any]):
        super().__init__(phenotype_converter, application_arguments)
        self._target_fitness = application_arguments[0]
        self._number_of_rounds = application_arguments[1]

    def calculate_fitness(self, genotype: Genotype) -> int:
        player = self._converter.convert(genotype)
        table = BlackJackTable([player])
        table.play_rounds(self._number_of_rounds)
        return player.money


def make_FitnessCalculator(application: Application, parameters: List[any]) -> FitnessCalculator:
    if application == Application.STRING_MATCH:
        if not isinstance(parameters[0], str):
            raise InvalidParameterException("First Parameter must be a str.")
        target_length = len(parameters[0])
        return FitnessCalculatorStringMatch(StringPhenotypeConverter(target_length), parameters)
    elif application == Application.PRIMES:
        if not isinstance(parameters[0], int):
            raise InvalidParameterException("First Parameter must be an int.")
        target_num_primes = parameters[0]  # how many primes we are trying to generate
        return FitnessCalculatorPrimes(target_num_primes, NumberGeneratorPhenotypeConverter(), parameters)
    elif application == Application.BLACK_JACK:
        if not isinstance(parameters[0], int):
            raise InvalidParameterException("Please specify how much total winnings is the goal.")
        if not isinstance(parameters[1], int):
            raise InvalidParameterException("Please specify how many rounds will be given to each player")
        return FitnessCalculatorBlackJack(PlayerConverter(), parameters)
    else:
        raise InvalidApplicationException("{} is not a valid choice".format(str(application)))


def make_PopulationFitnessCalculator(application: Application, parameters: List[any]) -> PopulationFitnessCalculator:
    if application == Application.BLACK_JACK:
        if not isinstance(parameters[0], int):
            raise InvalidParameterException("Please specify how much total winnings is the goal.")
        if not isinstance(parameters[1], int):
            raise InvalidParameterException("Please specify how many rounds will be given to each player")
        return PopulationFitnessCalculatorBlackJack(target_fitness=parameters[0], number_of_rounds=parameters[1])
    else:
        raise InvalidApplicationException("{} is not a valid choice".format(str(application)))



from enum import Enum, auto
from typing import List
import random
import struct

from gp_framework.bytegenotype import ByteGenotype
from gp_framework.phenotype.phenotype import PhenotypeConverter


class Card(Enum):
    ace = auto()
    two = auto()
    three = auto()
    four = auto()
    five = auto()
    six = auto()
    seven = auto()
    eight = auto()
    nine = auto()
    ten = auto()
    jack = auto()
    queen = auto()
    king = auto()


value_dict = {Card.two: 2, Card.three: 3, Card.four: 4, Card.five: 5, Card.six: 6, Card.seven: 7, Card.eight: 8,
              Card.nine: 9, Card.ten: 10, Card.jack: 10, Card.queen: 10, Card.king: 10}


class Deck:
    def __init__(self):
        """
        gives the instance a shuffled list of Cards
        """
        self._cards = [Card.ace, Card.two, Card.three, Card.four, Card.five, Card.six, Card.seven, Card.eight,
                       Card.nine, Card.ten, Card.jack, Card.queen, Card.king] * 4
        random.shuffle(self._cards)

    def take_random(self) -> Card:
        """
        :return: a card from deck or None if the deck is empty
        """
        if len(self._cards) == 0:
            return None
        card = self._cards[-1]
        del self._cards[-1]
        return card

    def __len__(self):
        return len(self._cards)


class PlayerStack:
    def __init__(self, starting_cards: List[Card]):
        self._cards = starting_cards

    def __len__(self):
        return len(self._cards)

    def add_card(self, card: Card):
        self._cards.append(card)

    @staticmethod
    def _normalize_score(score) -> int:
        if score > 21:
            return -1
        return score

    @staticmethod
    def _calc_possible_ace_scores(num_aces) -> List[int]:
        if num_aces == 0:
            return []
        possible_scores = [1, 11]
        num_aces -= 1
        for i in range(num_aces):
            possible_scores *= 2  # make a copy of each element
            possible_scores.sort()  # sort the list so that 1 doesn't always add up with 1, 11 with 11, etc
            for j in range(0, len(possible_scores), 2):
                possible_scores[j] += 1
                possible_scores[j + 1] += 11
            # remove duplicates and obvious bad choices
            possible_scores = set(possible_scores)
            for score in possible_scores:
                if score > 21:
                    del score
            possible_scores = list(possible_scores)
        return possible_scores

    @staticmethod
    def _maximize_score(num_aces, preliminary_score) -> int:
        # if there are no aces, return the normalized preliminary score
        if num_aces == 0:
            return PlayerStack._normalize_score(preliminary_score)

        possible_scores = PlayerStack._calc_possible_ace_scores(num_aces)
        for i in range(len(possible_scores)):
            possible_scores[i] = PlayerStack._normalize_score(possible_scores[i] + preliminary_score)
        return max(possible_scores)

    @property
    def score(self) -> int:
        num_aces = 0
        preliminary_score = 0
        for card in self._cards:
            if card == Card.ace:
                num_aces += 1
            else:
                preliminary_score += value_dict[card]
        return PlayerStack._maximize_score(num_aces, preliminary_score)


class PlayerParameters:
    """
    Everything that a Genotype decides about a player.
    """

    def __init__(self, score_weight: float, house_score_weight: float,pgw_weight: float, money_weight: float):
        self.score_weight = score_weight
        self.house_score_weight = house_score_weight
        self.pgw_weight = pgw_weight
        self.money_weight = money_weight

    @staticmethod
    def number_of_parameters() -> int:
        return 4

    @staticmethod
    def from_list(parameter_list: List[float]):
        """
        :param parameter_list: the values for each parameter in PlayerParameters
        :return: a PlayerParameters object
        """
        score_weight = parameter_list[0]
        house_score_weight = parameter_list[1]
        pgw_weight = parameter_list[2]
        money_weight = parameter_list[3]
        return PlayerParameters(score_weight, house_score_weight, pgw_weight, money_weight)


class Player:
    """
    Plays a game of Black Jack
    """
    _HIT_THRESHOLD = 1.0

    def __init__(self, parameters: PlayerParameters):
        self._parameters = parameters
        self._cards = None
        self.reset_cards()
        self._money = 100
        self._games_played = 0
        self._games_won = 0

    def hit(self, house_score: int) -> bool:
        hs_pressure = house_score * self._parameters.house_score_weight
        score_pressure = self._cards.score * self._parameters.score_weight
        return hs_pressure + score_pressure > Player._HIT_THRESHOLD

    def make_bet(self) -> int:
        """
        automatically decreases the Player's money
        :return: the player's starting bet
        """
        if self._games_played == 0:  # avoid div by zero errors
            self._money -= 10
            return 10
        float_bet = min(float(self._money),
                        self._money * self._parameters.money_weight + self.perc_games_won * self._parameters.pgw_weight)
        # converting to an integer like this can result in an OverflowError
        try:
            bet = int(float_bet)
        except OverflowError:
            if float_bet > 0:
                bet = self._money
            else:
                bet = 0

        bet = max(10, bet)
        self._money -= bet
        return bet

    def receive_card(self, card: Card):
        self._cards.add_card(card)

    def reset_cards(self):
        self._cards = PlayerStack([])

    def get_game_results(self, won: bool, winnings: int):
        self._games_played += 1
        if won:
            self._games_won += 1
        self._money += winnings

    @property
    def score(self):
        return self._cards.score

    @property
    def money(self):
        return self._money

    @property
    def perc_games_won(self):
        return self._games_won / self._games_played


class Dealer:
    """
    Acts as the house
    """

    def __init__(self):
        self._deck = Deck()
        hidden_card = self._deck.take_random()
        shown_card = self._deck.take_random()
        self._hidden_stack = PlayerStack([hidden_card, shown_card])
        self._shown_stack = PlayerStack([shown_card])

    def give_card(self) -> Card:
        return self._deck.take_random()

    def give_self_card(self) -> bool:
        """
        :return: whether or not the Dealer gave itself a new card
        """
        if self._hidden_stack.score >= 17:
            return False
        new_card = self._deck.take_random()
        if new_card is None:
            return False
        self._hidden_stack.add_card(new_card)
        self._shown_stack.add_card(new_card)

    @property
    def known_score(self):
        return self._shown_stack.score

    @property
    def hidden_score(self):
        return self._hidden_stack.score


class BlackJackTable:
    def __init__(self, players: List[Player]):
        """
        Important! Do NOT change the ordering of the list of players. It will mess up the mapping of Genotype to Money.
        This really should be fixed.
        """
        self._players = players
        self._payout = 3 / 2

    def _play_round(self):
        # set up
        dealer = Dealer()
        bets = {}
        for player in self._players:
            player.reset_cards()
            bets[player] = player.make_bet()
            # print("A bet of", bets[player], "has been placed.")
            player.receive_card(dealer.give_card())
            player.receive_card(dealer.give_card())

        # play
        cards_given = 0
        for player in self._players:
            while player.hit(dealer.known_score) and player.score > 0:
                new_card = dealer.give_card()
                """if new_card is None:
                    print("Warning! Deck is empty.")
                else:
                    cards_given += 1
                    print("Given {} cards".format(cards_given))"""
                player.receive_card(new_card)
            # print("Player done.\n")

        # payout
        for player in self._players:
            won = False
            winnings = 0
            player_score = player.score
            dealer_score = dealer.hidden_score
            # print("Player score: {} vs Dealer score {}".format(player_score, dealer_score))
            if player.score >= dealer.hidden_score:
                won = True
                winnings = bets[player] * self._payout
                # print("Paying out", winnings)
            player.get_game_results(won, winnings)

    def play_rounds(self, num_rounds):
        for i in range(num_rounds):
            self._play_round()

    @property
    def players(self) -> List[Player]:
        return self._players


class PlayerConverter(PhenotypeConverter):
    def convert(self, genotype: ByteGenotype) -> Player:
        """
        Turn genome into an array of parameters between 0 and 1 to be plugged into
        some application.

        :param genotype: The Genotype to convert
        :return: An array of floats between 0 and 1
        """
        parameters: List[float] = []

        # Each parameter will consume 4 bytes.
        for i in range(PlayerParameters.number_of_parameters()):
            [parameter] = struct.unpack('f', bytes(genotype[i:i + 4]))
            parameters.append(parameter)

        return Player(PlayerParameters.from_list(parameters))



from enum import Enum
from abc import ABC, abstractmethod

from gp_framework.bytegenotype import ByteGenotype


class Phenotype(Enum):
    STRING = "string"
    PARAMETERS = "parameters"


class PhenotypeConverter(ABC):
    @abstractmethod
    def convert(self, genotype: ByteGenotype) -> any:
        """
        Convert the given genotype to it's corresponding phenotype corresponding
        to the arguments passed on construction
        :param genotype:
        :return: Phenotype to be determined by concrete implementation.
        """
        pass


class ParametersPhenotypeConverter(PhenotypeConverter):
    def __init__(self, number_of_parameters):
        self._number_of_parameters = number_of_parameters

    def convert(self, genotype: ByteGenotype):
        """
        Turn genome into an array of parameters between 0 and 1 to be plugged into
        some application.

        :param genotype: The Genotype to convert
        :return: An array of floats between 0 and 1
        """
        parameters = []

        index_of_genome = 0
        for i in range(self._number_of_parameters):
            # Each parameter will consume 8 bytes, though the last 1 or 1 1/2 will
            # be lost to rounding. If the entire genome is used before finishing
            # the parameters, the function will circle around to the beginning of
            # the genome.
            parameter_to_add = "0x0."
            for j in range(8):
                if index_of_genome == len(genotype):
                    index_of_genome = 0
                parameter_to_add += str(hex(genotype[index_of_genome]))[2:]
                index_of_genome += 1
            parameter_to_add += "p0"
            # parameter_to_add should be a string "0x0.****************p0"
            parameters.append(float.fromhex(parameter_to_add))

        return parameters



from gp_framework import fitness_calculator

# Using Test Framework pytest

# TODO: Many of the random generation function should theoretically produce an
# even distribution of values between some min and max. How can we test for this.

def test_generate_random_genotype_single_byte_range():
    # Stress test to ensure generate_random_genotype() throws no errors

    # May print if desired to verify appropriate spread of integers between 0 and 255
    PRINTING = True
    for i in range(200):
        result = fitness_calculator.generate_random_genotype(1)
        if PRINTING:
            print(result[0])


def test__convert_to_parameters():
    genotype = fitness_calculator.generate_random_genotype(64)
    arguments = [3]
    print()
    print(fitness_calculator._convert_to_parameters(genotype, arguments))


def test__convert_to_string():
    genotype = fitness_calculator.generate_random_genotype(64)
    arguments = [11]
    print()
    print(fitness_calculator._convert_to_string(genotype, arguments))


def test_mutatate_genotype():
    # Should be able to observe infrequent bit flips in the genome
    PRINTING = True
    genotype = bytes([0b00000000, 0b00000000, 0b00000000, 0b11111111, 0b11111111, 0b11111111])
    mutation_factor = .1

    if PRINTING:
        print()
        print(genotype.hex())
    for i in range(25):
        genotype = fitness_calculator.mutate_genotype(genotype, mutation_factor)
        if PRINTING:
            print(genotype.hex())


def test__string_match_fitness():
    phenotype = "hello world"
    target_string = "hello world"
    perfect_fitness = fitness_calculator._string_match_fitness(phenotype, [target_string])
    print()
    print(perfect_fitness)

    phenotype = "hello worlc"
    fitness = fitness_calculator._string_match_fitness(phenotype, [target_string])
    print(fitness)
    assert fitness == perfect_fitness - 1

    phenotype = "hello wormd"
    fitness = fitness_calculator._string_match_fitness(phenotype, [target_string])
    print(fitness)
    assert fitness == perfect_fitness - 1

    phenotype = "           "
    fitness = fitness_calculator._string_match_fitness(phenotype, [target_string])
    print(fitness)


from gp_framework.fitness_calculator import Application


class Config:
    def __init__(self, bit_string_length: int, size_of_genotype: int, application: Application, show_plots: bool,
                 save_plots: bool):
        self._bit_string_length = bit_string_length
        self._size_of_genotype = size_of_genotype
        self._application = application
        self._show_plots = show_plots
        self._save_plots = save_plots

    @property
    def bit_string_length(self) -> int:
        return self._bit_string_length

    @property
    def size_of_genotype(self) -> int:
        return self._size_of_genotype

    @property
    def application(self) -> Application:
        return self._application

    @property
    def show_plots(self) -> bool:
        return self._show_plots

    @property
    def save_plots(self):
        return self._save_plots


CONFIG = Config(50, 11, Application.STRING_MATCH, True, True)



from gp_framework.phenotype.phenotype import PhenotypeConverter
from gp_framework.bytegenotype import ByteGenotype
from gp_framework.exception import InvalidParameterException


class StringPhenotypeConverter(PhenotypeConverter):
    def __init__(self, str_length: int):
        self._str_length = str_length

    def convert(self, genotype: ByteGenotype):
        """
        Convert a genotype to a string of ASCII characters

        :param genotype:
        :return: A string of the indicated length  consisting of each byte converted
         to an ASCII character
        """
        if len(genotype) < self._str_length:
            raise InvalidParameterException("Given genotype must be at least as "
                                            "long as the string being generated")

        result = ""

        for i in range(self._str_length):
            ascii_value = genotype[i]
            result += chr(ascii_value)

        return result



from gp_framework.evolutionary_optimizer import EvolutionaryOptimizer
from gp_framework.bytegenotype import ByteGenotype
from gp_framework import report as rep, meiosis, mitosis
from gp_framework.population_manager import *
from gp_framework.fitness_calculator import make_FitnessCalculator, Application
from gp_framework.phenotype.blackjack import PlayerParameters
from math import ceil, sqrt


def is_prime(x: int) -> bool:
    for i in range(2, ceil(sqrt(x))):
        if x % i == 0:
            return False
    return True


def main():
    # string_to_find = "hello world"
    # number_of_primes = 50
    # size_of_primes_genotype = 16
    black_jack_genotype_length = PlayerParameters.number_of_parameters() * 4  # 4 bytes per float
    fitness_calculator = make_FitnessCalculator(Application.BLACK_JACK, [1000, 20])
    population = ByteGenotype.generate_random_population(20, black_jack_genotype_length)

    simple_manager = mitosis.SimpleManager(population, fitness_calculator, "simple_mitosis_manager")
    truncation_manager = mitosis.TruncationManager(population, fitness_calculator, "truncation_manager")
    brute_force_manager = mitosis.BruteForce(population, fitness_calculator, "brute_force_manager")
    simple_meiosis_manager = meiosis.SimpleManager(population, fitness_calculator, "simple_meiosis_manager")
    multiple_children_manager = meiosis.MultipleChildrenManager(population, fitness_calculator, "multiple_children_manager")
    diversity_manager = meiosis.DiversityManager(population, fitness_calculator, "diversity_manager")
    tournament_manager = meiosis.TournamentManager(population, fitness_calculator, "tournament_manager")
    managers = [simple_meiosis_manager, multiple_children_manager, diversity_manager, brute_force_manager]

    optimizer = EvolutionaryOptimizer(managers, True)
    name_to_reports = optimizer.run_many_lifecycles(5000)
    rep.generate_many_reports(LifecycleReport.header(), name_to_reports, {}, 1)
    solutions: List[Tuple[str, any]] =\
        [(elem[0], fitness_calculator.converter.convert(elem[1][-1].solution)) for elem in name_to_reports.items()]

    print("Found solutions:")
    for elem in solutions:
        print("From", elem[0])
        print("Generated word:", elem[1])


main()



from gp_framework.genotype import Genotype
from random import randrange, random


class ByteGenotype(Genotype):
    def __init__(self, array_of_bytes: bytearray):
        """
        Initialize an instance of Genotype
        The hashcode is set here because the hashcode must remain unchanged once an object is in a set.
        :param array_of_bytes: The underlying representation of the genotype. For some
                application, each byte can represent an ascii character
        """
        self._array_of_bytes = array_of_bytes
        hashcode = 0
        for byte in array_of_bytes:
            hashcode += byte
        self._hashcode = hashcode

    def __getitem__(self, item):
        return self._array_of_bytes[item]

    def __len__(self):
        return len(self._array_of_bytes)

    def _mutate_bytearray(self, mutation_factor) -> bytearray:
        """
        :param mutation_factor: the probability of having a 1 at any given index in the bitmask should be in [0.0, 1.0]
        """
        # newbyte = byte ^ (xor) mask -> this will flip the bits that are on in the mask.

        result = []
        for byte in self._array_of_bytes:

            # Create a mask for each byte
            bitmask = 0
            if random() < mutation_factor:
                bitmask += 1
            for i in range(7):
                bitmask = bitmask << 1
                if random() < mutation_factor:
                    bitmask += 1

            result.append(byte ^ bitmask)

        # Convert list of bytes to type bytes
        return bytearray(result)

    def mutate(self, mutation_factor: float) -> None:
        self._array_of_bytes = self._mutate_bytearray(mutation_factor)

    def make_mutated_copy(self, mutation_factor):
        return ByteGenotype(self._mutate_bytearray(mutation_factor))

    def __eq__(self, other):
        if not isinstance(other, ByteGenotype):
            return False
        return self._array_of_bytes == other._array_of_bytes

    def __ne__(self, other):
        return not self.__eq__(other)

    def __hash__(self):
        return self._hashcode

    @staticmethod
    def generate_random_genotype(genotype_size: int):
        """
        Factory method to randomly generate an instance of Genotype. For some
        application, each byte can represent an ascii character.
        :param genotype_size: Desired length of the returned Genotype
        :return: Randomly generated Genotype
        """

        array = bytearray()
        for byte in range(genotype_size):
            # Generate byte, plug mask into each bit
            new_byte = randrange(2)
            for bit in range(7):
                new_byte = new_byte << 1
                if 1 == randrange(2):
                    new_byte += 1

            array.append(new_byte)

        return ByteGenotype(array)

    @staticmethod
    def generate_random_population(population_size, genotype_size):
        """
        Convenience method to generate a list of random genotypes
        :param population_size: length of list to return
        :param genotype_size: size of each genotype
        :return: list of Genotypes
        """

        return [ByteGenotype.generate_random_genotype(genotype_size) for _ in range(population_size)]



"""
Mitosis is splitting DNA strands and mutating them
"""
from operator import itemgetter
from gp_framework.population_manager import *
from gp_framework.bytegenotype import ByteGenotype


class SimpleManager(PopulationManager):
    def produce_offspring(self, population: List[ByteGenotype]) -> Tuple[List[ByteGenotype], List[ByteGenotype]]:
        judged_population = self.calculate_population_fitness(population)
        fittest_individual_tuple = max(judged_population, key=itemgetter(1))
        fitness: float = fittest_individual_tuple[1]
        children = []

        for _ in range(len(population)):
            child: ByteGenotype = fittest_individual_tuple[0]
            child.mutate((1.0 - fitness) * .0001)
            children.append(child)

        return [fittest_individual_tuple[0]], children

    def select_next_generation(self, parents: List[ByteGenotype], children: List[ByteGenotype]) -> List[ByteGenotype]:
        self._newest_report = self.make_LifecycleReport(children)
        return children


class TruncationManager(PopulationManager):
    def produce_offspring(self, population: List[ByteGenotype]) -> Tuple[List[ByteGenotype], List[ByteGenotype]]:
        children = []
        for genotype in population:
            child = genotype
            child.mutate(.0001)
            children.append(child)
        return population, children

    def select_next_generation(self, parents: List[ByteGenotype], children: List[ByteGenotype]) -> List[ByteGenotype]:
        all_genotypes = self.calculate_population_fitness(parents + children)
        all_genotypes.sort(key=lambda e: e[1], reverse=True)
        half_way = int(len(all_genotypes) / 2)
        new_population = [g[0] for g in all_genotypes[0:half_way]]
        self._newest_report = self.make_LifecycleReport(new_population)
        return new_population


class BruteForce(PopulationManager):
    def __init__(self, population: List[ByteGenotype], fitness_calculator: FitnessCalculator, name: str = "Brute Force Manager"):
        super().__init__(population, fitness_calculator, name)
        judged_population = self.calculate_population_fitness(population)
        judged_population.sort(key=lambda e: e[1], reverse=True)
        self._best_genotype = judged_population[0][0]
        self._best_fitness = -1

    def produce_offspring(self, population: List[ByteGenotype]) -> Tuple[List[ByteGenotype], List[ByteGenotype]]:
        children = ByteGenotype.generate_random_population(20, len(self._best_genotype))
        return [self._best_genotype], children

    def select_next_generation(self, parents: List[ByteGenotype], children: List[ByteGenotype]) -> List[ByteGenotype]:
        combined_pop_with_fitness = self.calculate_population_fitness(parents + children)
        combined_pop_with_fitness.sort(key=lambda elem: elem[1], reverse=True)
        self._best_genotype = combined_pop_with_fitness[0][0]
        self._best_fitness = combined_pop_with_fitness[0][1]
        # this could just be a list with just the fittest individual since that's what'll be output anyways
        self._newest_report = self.make_LifecycleReport([elem[0] for elem in combined_pop_with_fitness])
        return []



import unittest

from gp_framework import report as rep


class TestReportModule(unittest.TestCase):
    def test_transpose_list_of_lists(self):
        input_list = [['a', 'b', 'c', 'd', 'e'], ['f', 'g', 'h', 'i', 'j'], ['k', 'l', 'm', 'n', 'o']]
        expected_output = [['a', 'f', 'k'], ['b', 'g', 'l'], ['c', 'h', 'm'], ['d', 'i', 'n'], ['e', 'j', 'o']]
        actual_output = rep._transpose_list_of_lists(input_list)
        self.assertEqual(actual_output, expected_output)

    def test_combine_list_elements_group_size_2(self):
        input_list = [i*2 for i in range(10)]
        expected_output = [1, 5, 9, 13, 17]
        actual_output = rep._combine_list_elements(input_list, 2)
        self.assertEqual(actual_output, expected_output)

    def test_combine_list_elements_group_size_3(self):
        input_list = [i for i in range(10)]
        expected_output = [1, 4, 7, 9]
        actual_output = rep._combine_list_elements(input_list, 3)
        self.assertEqual(actual_output, expected_output)


if __name__ == '__main__':
    unittest.main()



from abc import ABC, abstractmethod


class Genotype(ABC):
    @abstractmethod
    def mutate(self, mutation_factor: float):
        pass

    @abstractmethod
    def make_mutated_copy(self, mutation_factor: float):
        pass

    @abstractmethod
    def __getitem__(self, item):
        pass

    @abstractmethod
    def __len__(self):
        pass

    @abstractmethod
    def __eq__(self, other):
        pass

    @abstractmethod
    def __ne__(self, other):
        pass

    @abstractmethod
    def __hash__(self):
        pass



from typing import List, Tuple, Collection, Set
from abc import abstractmethod, ABC

from gp_framework.fitness_calculator import FitnessCalculator
from gp_framework.bytegenotype import Genotype


class LifecycleReport:
    """
    It's a POJO for whatever data we think is good to keep track of from generation to generation
    """

    def __init__(self, max_fitness=-1.0, diversity=-1.0, solution: Genotype = None):
        """
        :param max_fitness: the fitness of the most fit genotype of this generation
        :param diversity: a measure from 0.0 to 1.0 of how many unique individuals there are in the population
        :param solution: the most fit genotype of this generation
        """
        self._max_fitness = max_fitness
        self._diversity = diversity
        self._solution = solution

    def to_list(self):
        """
        useful for making csv files
        """
        return [self.max_fitness, self.diversity]

    @staticmethod
    def header() -> List[str]:
        """
        useful for making csv files
        """
        return ['max_fitness', 'diversity']

    @property
    def max_fitness(self):
        return self._max_fitness

    @property
    def diversity(self):
        return self._diversity

    @property
    def solution(self) -> Genotype:
        return self._solution


class PopulationManager(ABC):
    """
    This abstract class wraps up useful default behavior for searching the solution space.
    Subclass and override the behavior in produce_offspring and select_next_generation. Then, just call lifecycle.
    """

    def __init__(self, population: List[Genotype],
                 fitness_calculator: FitnessCalculator, name: str = "Unnamed PopulationManager"):
        """
        :param population: The starting population
        calculator, this converts the provided genotype to a phenotype accepted
        by the fitness_calculator.
        :param fitness_calculator: This is used to judge our solutions
        :param name: the name that this PopulationManager will be known as in reports
        """
        self._population = population
        self._fitness_calculator = fitness_calculator
        # this should be set in produce_offspring or select_next_generation and is returned by lifecycle
        self._newest_report: LifecycleReport = LifecycleReport()
        self._name = name
        self._best_genotypes: Set[Genotype] = set()
        self._best_genotype_fitness = -1

    @property
    def name(self) -> str:
        return self._name

    @property
    def best_genotypes(self):
        return self._best_genotypes

    @property
    def best_genotype_fitness(self):
        return self._best_genotype_fitness

    def make_LifecycleReport(self, genotypes: Collection[Genotype]) -> LifecycleReport:
        max_fitness = -1
        fittest_genotype = None

        for genotype in genotypes:
            fitness = self._fitness_calculator.calculate_fitness(genotype)
            if fitness > max_fitness:
                max_fitness = fitness
                fittest_genotype = genotype

        unique_genotypes = set(genotypes)
        diversity = len(unique_genotypes) / len(genotypes)

        return LifecycleReport(max_fitness, diversity, fittest_genotype)

    def calculate_population_fitness(self, population: Collection[Genotype]) -> List[Tuple[Genotype, float]]:
        """
        Use fitness_calculator to assign a rank to each Genotype in the population
        :return: each member of the population with their fitness, a summary of important findings
        """

        judged_population: List[Tuple[Genotype, float]] = []

        for genotype in population:
            fitness = self._fitness_calculator.calculate_normalized_fitness(genotype)
            judged_population.append((genotype, fitness))

        return judged_population

    @abstractmethod
    def produce_offspring(self, population: List[Genotype]) -> Tuple[List[Genotype], List[Genotype]]:
        """
        Create new Genotypes from the old ones
        :param population: A list of tuples pairing a Genotype with its fitness
        :return: The parents (selected from population) who created the offspring, The offspring of population
        """
        pass

    @abstractmethod
    def select_next_generation(self, parents: List[Genotype], children: List[Genotype]) -> List[Genotype]:
        """
        Select M individuals to make up the next generation of Genotypes
        :param parents: The old generation paired with their fitnesses
        :param children: The offspring of parents
        :return: the new generation
        """
        pass

    def lifecycle(self) -> LifecycleReport:
        """
        Run the population of solutions through a selection process. self._population is not updated until the end
        of this method and can thus be accessed in produce_offspring and select_next_generation.
        :return: a summary of important findings
        """

        parents, children = self.produce_offspring(self._population)
        self._population = self.select_next_generation(parents, children)

        # update
        if self._newest_report.max_fitness > self._best_genotype_fitness:
            self._best_genotype_fitness = self._newest_report.max_fitness
            self._best_genotypes = {self._newest_report.solution}
        elif self._newest_report.max_fitness == self._best_genotype_fitness:
            self._best_genotypes.add(self._newest_report.solution)

        return self._newest_report



"""
Meiosis is mixing bits of DNA from parents
"""
import math

from gp_framework.population_manager import PopulationManager
from gp_framework.fitness_calculator import StringPhenotypeConverter, FitnessCalculator
from gp_framework.bytegenotype import ByteGenotype
from random import choice, randint
from typing import List, Tuple

MUTATION_RATE = .02


class SimpleManager(PopulationManager):
    """
    Takes a very simplistic approach to creating new Genotypes from existing ones
    """
    def produce_offspring(self, population: List[ByteGenotype]) -> Tuple[List[ByteGenotype], List[ByteGenotype]]:
        judged_population = self.calculate_population_fitness(population)
        judged_population.sort(key=lambda e: e[1], reverse=True)  # it's still in tuple form
        fittest_individual: ByteGenotype = judged_population[0][0]

        children = []

        for genotype in population:
            children.append(self.create_genotype([genotype, fittest_individual]))

        return population, children

    @staticmethod
    def create_genotype(parents: List[ByteGenotype]) -> ByteGenotype:
        """
        Randomly select bits of the parent Genotypes to create a new child Genotype
        :param parents: A list of all the possible parents for the child Genotype
        """
        child_byte_array = bytearray()
        for i in range(len(parents[0])):
            parent = choice(parents)
            child_byte_array.append(parent[i])
        child = ByteGenotype(child_byte_array)
        child.mutate(MUTATION_RATE)
        return child

    def select_next_generation(self, parents: List[ByteGenotype], children: List[ByteGenotype]) -> List[ByteGenotype]:
        self._newest_report = self.make_LifecycleReport(children)
        return children


class MultipleChildrenManager(PopulationManager):
    """
    Each set of parent Genotypes produces multiple child Genotypes
    """
    def __init__(self, population: List[ByteGenotype], fitness_calculator: FitnessCalculator, name: str):
        super().__init__(population, fitness_calculator, name)
        # M is the population size. This variable is used to make sure the actual population size stays constant
        self._M = len(population)
        self._num_children = 3

    def produce_offspring(self, population: List[ByteGenotype]) -> Tuple[List[ByteGenotype], List[ByteGenotype]]:
        judged_population = self.calculate_population_fitness(population)
        judged_population.sort(key=lambda e: e[1], reverse=True)  # it's still in tuple form
        fittest_individual: ByteGenotype = judged_population[0][0]

        children = []

        for genotype in population:
            children += self.produce_multiple_children([fittest_individual, genotype, choice(population)])

        return population, children

    def produce_multiple_children(self, parents: List[ByteGenotype]) -> List[ByteGenotype]:
        children = []
        for i in range(self._num_children):
            children.append(MultipleChildrenManager.create_genotype(parents))
        return children

    @staticmethod
    def create_genotype(parents: List[ByteGenotype]) -> ByteGenotype:
        """
        Randomly select bits of the parent Genotypes to create a new child Genotype
        :param parents: A list of all the possible parents for the child Genotype
        """
        child_byte_array = bytearray()
        for i in range(len(parents[0])):
            parent = choice(parents)
            child_byte_array.append(parent[i])
        child = ByteGenotype(child_byte_array)
        child.mutate(MUTATION_RATE)
        return child

    def select_next_generation(self, parents: List[ByteGenotype], children: List[ByteGenotype]) -> List[ByteGenotype]:
        combined_population = parents + children
        unique_population = set(combined_population)
        # print('# unique Genotypes:', len(unique_population))
        judged_population = self.calculate_population_fitness(unique_population)
        judged_population.sort(key=lambda e: e[1], reverse=True)

        next_generation = [judged_population[i][0] for i in range(self._M)]
        # Make LifecycleReport with combined_population so that we actually know what the diversity is
        self._newest_report = self.make_LifecycleReport(combined_population)
        return next_generation


class DiversityManager(PopulationManager):
    """
    Encourages diversity in the population
    """
    def __init__(self, population: List[ByteGenotype], fitness_calculator: FitnessCalculator, name: str):
        super().__init__(population, fitness_calculator, name)
        # M is the population size. This variable is used to make sure the actual population size stays constant
        self._M = len(population)
        self._num_children = 2

    def produce_offspring(self, population: List[ByteGenotype]) -> Tuple[List[ByteGenotype], List[ByteGenotype]]:
        judged_population = self.calculate_population_fitness(population)
        judged_population.sort(key=lambda e: e[1], reverse=True)  # it's still in tuple form
        fittest_individual: ByteGenotype = judged_population[0][0]

        children = []

        for genotype in population:
            children += self.produce_multiple_children([fittest_individual, genotype])

        return population, children

    def produce_multiple_children(self, parents: List[ByteGenotype]) -> List[ByteGenotype]:
        children = []
        for i in range(self._num_children):
            children.append(MultipleChildrenManager.create_genotype(parents))
        return children

    @staticmethod
    def create_genotype(parents: List[ByteGenotype]) -> ByteGenotype:
        """
        Randomly select bits of the parent Genotypes to create a new child Genotype
        :param parents: A list of all the possible parents for the child Genotype
        """
        child_byte_array = bytearray()
        for i in range(len(parents[0])):
            parent = choice(parents)
            child_byte_array.append(parent[i])
        child = ByteGenotype(child_byte_array)
        child.mutate(MUTATION_RATE)
        return child

    def _choose_genotypes_from_sorted_unique_list(self, genotypes: List[ByteGenotype]) -> List[ByteGenotype]:
        """
        Takes in all the unique Genotypes from the population and selects some to keep
        :param genotypes: A sorted list of unique Genotypes
        """
        splitting_point = math.floor(self._M/2)
        good = [genotypes[i] for i in range(self._M - splitting_point)]
        # bad = [genotypes[i] for i in range(-1, -splitting_point-1, -1)]
        bad = ByteGenotype.generate_random_population(splitting_point, len(genotypes[0])) #insert garbage Genotypes
        return good + bad

    def select_next_generation(self, parents: List[ByteGenotype], children: List[ByteGenotype]) -> List[ByteGenotype]:
        combined_population = parents + children
        unique_population = set(combined_population)
        # print('# unique Genotypes:', len(unique_population))
        judged_population = self.calculate_population_fitness(unique_population)
        judged_population.sort(key=lambda e: e[1], reverse=True)

        # Make LifecycleReport with combined_population so that we actually know what the diversity is
        self._newest_report = self.make_LifecycleReport(combined_population)
        return self._choose_genotypes_from_sorted_unique_list([g[0] for g in judged_population])


class TournamentManager(PopulationManager):
    """
    Uses tournament selection
    """
    def produce_offspring(self, population: List[ByteGenotype]) -> Tuple[List[ByteGenotype], List[ByteGenotype]]:
        judged_population = self.calculate_population_fitness(population)
        parents = []
        for pair in judged_population:
            parents.append(max(pair, choice(judged_population), key=lambda t: t[1])[0]) # this is where the tournament happens
        children = self.create_offspring(parents)
        return parents, children

    def create_offspring(self, parents: List[ByteGenotype]) -> List[ByteGenotype]:
        children = []
        for _ in parents:
            children.append(self.crossover_and_mutate(choice(parents), choice(parents)))
        return children

    @staticmethod
    def crossover_and_mutate(parent1: ByteGenotype, parent2: ByteGenotype) -> ByteGenotype:
        child_array = []
        for i in range(len(parent1)):
            parent_choice = randint(0, 1)
            parent: ByteGenotype
            if parent_choice == 0:
                parent = parent1
            else:
                parent = parent2
            child_array.append(parent[i])
        child = ByteGenotype(bytearray(child_array))
        child.mutate(MUTATION_RATE)
        return child

    def select_next_generation(self, parents: List[ByteGenotype], children: List[ByteGenotype]) -> List[ByteGenotype]:
        self._newest_report = self.make_LifecycleReport(children)
        return children



import unittest

from gp_framework.phenotype.blackjack import PlayerStack, Card, Dealer


class ScoreTest(unittest.TestCase):
    def test_score_basic(self):
        stack = PlayerStack([Card.ten, Card.six])
        self.assertEqual(stack.score, 16)

    def test_score_bust(self):
        stack = PlayerStack([Card.queen, Card.jack, Card.two])
        self.assertEqual(stack.score, -1)

    def test_score_one_ace(self):
        stack = PlayerStack([Card.king, Card.ace])
        self.assertEqual(stack.score, 21)

    def test_score_two_aces(self):
        stack = PlayerStack([Card.ace, Card.ace, Card.nine])
        self.assertEqual(stack.score, 21)

    def test_score_three_aces(self):
        stack = PlayerStack([Card.ace, Card.ace, Card.ace])
        self.assertEqual(stack.score, 13)

    def test_score_four_aces(self):
        stack = PlayerStack([Card.ace, Card.ace, Card.ace, Card.ace, Card.nine])
        self.assertEqual(stack.score, 13)


class DealerTest(unittest.TestCase):
    def test_give_card(self):
        dealer = Dealer()
        # check that the dealer gave itself the correct amount of cards
        self.assertEqual(len(dealer._hidden_stack), 2)
        self.assertEqual(len(dealer._shown_stack), 1)
        self.assertEqual(len(dealer._deck), 50)

        cards = []
        for _ in range(50):
            cards.append(dealer.give_card())
        for elem in cards:
            self.assertIsInstance(elem, Card)
        should_be_none = dealer.give_card()
        self.assertIsNone(should_be_none)


if __name__ == '__main__':
    unittest.main()



from gp_framework.genotype import Genotype
from typing import List
import numpy as np


class FloatGenotype(Genotype):
    def __init__(self, floats: List[float]):
        self._floats = floats
        self._hashcode = None

    def _mutate_floats(self, mutation_factor: float) -> List[float]:
        return [f + np.random.uniform(-mutation_factor, mutation_factor, 1)[0] for f in self._floats]

    def mutate(self, mutation_factor: float):
        self._floats = self._mutate_floats(mutation_factor)

    def make_mutated_copy(self, mutation_factor):
        return FloatGenotype(self._mutate_floats(mutation_factor))

    def __getitem__(self, item):
        return self._floats[item]

    def __len__(self):
        return len(self._floats)

    def __eq__(self, other):
        if not isinstance(other, FloatGenotype):
            return False
        return self._floats == other._floats

    def __ne__(self, other):
        return not self.__eq__(other)

    def __hash__(self):
        if self._hashcode is None:
            self._hashcode = int(sum(self._floats) * 10)
        return self._hashcode

    @staticmethod
    def generate_random_genotype(genotype_size: int, lower_bound: float, upper_bound: float):
        floats = [np.random.uniform(lower_bound, upper_bound, 1)[0] for _ in range(genotype_size)]
        return FloatGenotype(floats)

    @staticmethod
    def generate_random_population(population_size: int, genotype_size: int, lower_bound: float, upper_bound: float):
        return [FloatGenotype.generate_random_genotype(genotype_size, lower_bound, upper_bound)
                for _ in range(population_size)]



import plotly.graph_objects as go
from plotly.subplots import make_subplots
import csv
from typing import Dict
from gp_framework.fitness_calculator import *
from gp_framework.population_manager import LifecycleReport
from gp_framework import config


_END_OF_METADATA = ["End of metadata"]


def generate_many_reports(header: List[str], name_to_reports: Dict[str, List[LifecycleReport]],
                          name_to_metadata: Dict[str, List[List[any]]], elements_per_point):
    for item in name_to_reports.items():
        # the [] in name_to_metadata.get is the value returned if item[0] is not a valid key
        generate_csv(item[0] + '.csv', header, [r.to_list() for r in item[1]], name_to_metadata.get(item[0], []))
        generate_plot_from_csv(item[0] + '.csv', elements_per_point, item[0])


def generate_csv(csv_name: str, header: List[any], rows: List[List[any]], metadata: List[List[any]]):
    with open("csvs/" + csv_name, 'w') as csv_file:
        csv_writer = csv.writer(csv_file, quoting=csv.QUOTE_NONNUMERIC)
        csv_writer.writerows(metadata)
        csv_writer.writerow(_END_OF_METADATA)
        csv_writer.writerow(header)
        csv_writer.writerows(rows)


def _average(floats: List[float]) -> float:
    total = 0
    for i in floats:
        total += i
    return total/len(floats)


def _combine_list_elements(list_: List[float], group_size: int) -> List[float]:
    combined_list = []
    starting_indices = [i for i in range(0, len(list_), group_size)]
    end_index = len(list_) + 1
    for index in starting_indices:
        combined_list.append(_average(list_[index:min(index+group_size, end_index)]))
    return combined_list


def _transpose_list_of_lists(list_of_lists: List[List[any]]) -> List[List[any]]:
    """
    This assumes that all inner_lists have the same length
    :param list_of_lists: the list to transpose
    :return: the transposed list
    """
    new_list = []
    for i in range(len(list_of_lists[0])):
        new_list.append([])
        for j in range(len(list_of_lists)):
            new_list[i].append(list_of_lists[j][i])
    return new_list


def generate_plot_from_csv(csv_name: str, elements_per_point: int, output_name: str) -> None:
    """
    Makes nice plots to help visualize data
    :param csv_name: Name of csv file to draw data from
    :param elements_per_point: How many data points to average into one point on the plot
    :return:
    """

    labels: List[str]
    data: List[List[float]] = []

    # read the csv file into a list of lists
    with open("csvs/{}".format(csv_name), 'r') as file:
        reader = csv.reader(file, quoting=csv.QUOTE_NONNUMERIC)

        # scan until the end of the metadata is reached
        for row in reader:
            if row == _END_OF_METADATA:
                break

        labels = next(reader)
        for row in reader:
            data.append(row)
    data = _transpose_list_of_lists(data)

    # combine the elements of data
    for i in range(len(data)):
        data[i] = _combine_list_elements(data[i], elements_per_point)

    # fig = make_subplots(rows=len(data), cols=1, subplot_titles=labels)
    fig = go.Figure()
    for i in range(len(data)):
        fig.add_trace(go.Scatter(x=[j for j in range(len(data[i]))], y=data[i], name=labels[i]))
    fig.update_layout(height=1000, width=1000*len(data), title_text=output_name)

    if config.CONFIG.save_plots:
        fig.write_html("plots/{}.html".format(output_name))
    if config.CONFIG.show_plots:
        fig.show()



from gp_framework.bytegenotype import ByteGenotype
from gp_framework.phenotype.phenotype import PhenotypeConverter


class NumberGenerator:
    def __init__(self, w: int, x: int, w_delta: int, x_delta: int):
        self._w = w
        self._x = x
        self._w_delta = w_delta
        self._x_delta = x_delta
        self._list_of_numbers = [2]

    def _add_to_list_of_primes(self):
        prime = self._list_of_numbers[-1] * self._w + self._x
        self._list_of_numbers.append((abs(prime) % 10_000) + 2)
        self._update_constants()

    def _update_constants(self):
        self._w += self._w_delta
        self._x += self._x_delta

    def make_list(self, size):
        for _ in range(size):
            self._add_to_list_of_primes()

    def __str__(self):
        return "w = {}; x = {}; w_delta = {}; x_delta = {}".format(
            self._w, self._x, self._w_delta, self._x_delta)

    @property
    def list_of_numbers(self):
        return self._list_of_numbers


class NumberGeneratorPhenotypeConverter(PhenotypeConverter):
    def convert(self, genotype: ByteGenotype) -> NumberGenerator:
        # assign each of these values based on the provided Genotype
        w: int
        x: int
        w_delta: int
        x_delta: int

        new_bytes = []
        i = 0
        while len(new_bytes) < 16:  # 16 because we need 4 bytes for each of the desired ints
            new_bytes.append(genotype[i % len(genotype)])
            i += 1

        w = self._to_int(new_bytes[:4])
        x = self._to_int(new_bytes[4:8])
        w_delta = self._to_int(new_bytes[8:12])
        x_delta = self._to_int(new_bytes[12:16])
        return NumberGenerator(w, x, w_delta, x_delta)

    @staticmethod
    def _to_int(bytes_) -> int:
        return int.from_bytes(bytes_, byteorder="little", signed=True)



import time
from typing import List, Dict
from gp_framework.population_manager import PopulationManager, LifecycleReport
from gp_framework.fitness_calculator import NumberGeneratorPhenotypeConverter, NumberGenerator, StringPhenotypeConverter


class EvolutionaryOptimizer:
    def __init__(self, managers: List[PopulationManager], show_output: bool):
        self._managers = managers
        self._show_output = show_output

    def _run_lifecycles(self, manager: PopulationManager, max_iterations: int = -1) -> List[LifecycleReport]:
        if self._show_output:
            print("Began {} at {}.".format(manager.name, _time()))

        i = 0
        reports = []
        # todo decide how to implement metadata, maybe have a list inside LifecycleReport
        while i < max_iterations and (len(reports) == 0 or reports[len(reports) - 1].max_fitness != 1.0):
            reports.append(manager.lifecycle())
            i += 1

        if self._show_output:
            print("Finished {} ({} iterations) at {}.".format(manager.name, i, _time()))
            if reports[-1].solution is not None:
                converter = StringPhenotypeConverter(len(reports[-1].solution))
                print("Solution:", converter.convert(reports[-1].solution))
            print()

        return reports

    def run_many_lifecycles(self, max_iterations) -> Dict[str, List[LifecycleReport]]:
        manager_name_to_reports = {}
        for manager in self._managers:
            manager_name_to_reports[manager.name] = self._run_lifecycles(manager, max_iterations)

        return manager_name_to_reports


def _time() -> str:
    return time.asctime(time.localtime(time.time()))
